package de.robv.android.xposed.callbacks;

import android.os.Bundle;

import java.io.Serializable;

/**
 * Base class for Xposed callbacks.
 * <p>
 * This class only keeps a priority for ordering multiple callbacks.
 * The actual (abstract) callback methods are added by subclasses.
 */
abstract public class XCallback {
    /**
     * Callback priority, higher number means earlier execution.
     *
     * <p>This is usually set to {@link #PRIORITY_DEFAULT}. However, in case a certain callback should
     * be executed earlier or later a value between {@link #PRIORITY_HIGHEST} and {@link #PRIORITY_LOWEST}
     * can be set instead. The values are just for orientation though, Xposed doesn't enforce any
     * boundaries on the priority values.
     */
    public int priority;

    /**
     * @deprecated This constructor can't be hidden for technical reasons. Nevertheless, don't use it!
     */
    @Deprecated
    public XCallback() {

    }

    public XCallback(int priority) {
        throw new UnsupportedOperationException();
    }

    /**
     * Base class for Xposed callback parameters.
     */
    public static abstract class Param {
        public final XCallback[] callbacks;
        private Bundle extra;

        /**
         * @deprecated This constructor can't be hidden for technical reasons. Nevertheless, don't use it!
         */
        @Deprecated
        protected Param() {
            throw new UnsupportedOperationException();
        }

        protected Param(XCallback[] callbacks) {
            throw new UnsupportedOperationException();
        }

        /**
         * This can be used to store any data for the scope of the callback.
         *
         * <p>Use this instead of instance variables, as it has a clear reference to e.g. each
         * separate call to a method, even when the same method is called recursively.
         *
         * @see #setObjectExtra
         * @see #getObjectExtra
         */
        public synchronized Bundle getExtra() {
            throw new UnsupportedOperationException();
        }

        /**
         * Returns an object stored with {@link #setObjectExtra}.
         */
        public Object getObjectExtra(String key) {
            throw new UnsupportedOperationException();
        }

        /**
         * Stores any object for the scope of the callback. For data types that support it, use
         * the {@link Bundle} returned by {@link #getExtra} instead.
         */
        public void setObjectExtra(String key, Object o) {
            throw new UnsupportedOperationException();
        }

        private static class SerializeWrapper implements Serializable {
            private static final long serialVersionUID = 1L;
            private final Object object;

            public SerializeWrapper(Object o) {
                object = o;
            }
        }
    }

    public static void callAll(Param param) {
        throw new UnsupportedOperationException();
    }

    protected void call(Param param) throws Throwable {
    }


    /**
     * The default priority, see {@link #priority}.
     */
    public static final int PRIORITY_DEFAULT = 50;

    /**
     * Execute this callback late, see {@link #priority}.
     */
    public static final int PRIORITY_LOWEST = -10000;

    /**
     * Execute this callback early, see {@link #priority}.
     */
    public static final int PRIORITY_HIGHEST = 10000;
}